package enverstest.model;

import static org.junit.Assert.*;

import java.util.List;

import javax.persistence.EntityManager;

import org.hibernate.envers.AuditReader;
import org.hibernate.envers.AuditReaderFactory;
import org.hibernate.envers.query.AuditEntity;
import org.hibernate.envers.query.AuditQuery;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

import enverstest.util.JpaUtil;

public class CategoryTest
{
	private EntityManager em;
	
	@Before
	public void setUp() throws Exception 
	{
		em = JpaUtil.getEntityManager("enverstest");
	}

	@After
	public void tearDown() throws Exception 
	{
		JpaUtil.shutdown();
	}

	@Test
	public void testEnvers() 
	{
		long startPIT = System.currentTimeMillis();
		
		System.out.println("\n===== Fill Database =====\n");

		JpaUtil.beginTransaction();
		
		Item item = new Item("The Item");
		
		assertNull(item.getId());
		
		em.persist(item);
		
		assertNotNull(item.getId());
		
		Category category = new Category("The Category");
		
		category.setDescription("The description");
		
		category.setValue(item, new Value("The Value", 4711L));
		
		category.setText(item, "The Text");
		
		assertNull(category.getId());
		
		em.persist(category);
		
		assertNotNull(category.getId());
		
		JpaUtil.commitTransaction();
		
		System.out.println("\n===== Database filled =====\n");

		em.clear();
		
		System.out.println("\n===== Check current entities =====\n");
		
		JpaUtil.beginTransaction();
		
		Category loadedCategory = em.find(Category.class, category.getId());
		
		assertNotNull(loadedCategory);
		assertEquals("The Category", loadedCategory.getName());
		assertEquals("The description", loadedCategory.getDescription());
		
		Value theValue = loadedCategory.getValue(item);
		assertNotNull(theValue);
		assertEquals("The Value", theValue.getText());
		assertEquals(4711L, theValue.getNumber().longValue());
		
		assertEquals("The Text", loadedCategory.getText(item));
		
		JpaUtil.commitTransaction();
		
		System.out.println("\n===== Load historic data =====\n");
		
		JpaUtil.beginTransaction();
		
		// first we retrieve the first revision of the Category entity
		AuditReader reader = AuditReaderFactory.get(em);
		AuditQuery query = reader.createQuery().forRevisionsOfEntity(Category.class, false, true)
		                                       .addOrder(AuditEntity.revisionProperty("timestamp").asc())
		                                       .add(AuditEntity.id().eq(category.getId()));
		
		@SuppressWarnings("unchecked")
		List<Object[]> history = (List<Object[]>)query.getResultList();
		
		assertNotNull(history);
		assertEquals(1, history.size());
		
		Revision first = (Revision)history.get(0)[1];
		
		assertNotNull(first);
		assertTrue(startPIT < first.getTimestamp().longValue());
		
		// Now read historic data of the category entity at the first revision
		// This will of course be the same as the current entity, but it is enough
		// to show the problem
		query = reader.createQuery().forEntitiesAtRevision(Category.class, first.getId())
		                            .add(AuditEntity.property("id").eq(category.getId()))
		                            .setMaxResults(1);
		
		Category historicCategory = (Category)query.getSingleResult();
		
		assertNotNull(historicCategory);
		
		System.out.println("\n===== Check historic data =====\n");
		
		assertEquals(category.getId(), historicCategory.getId());
		assertEquals("The Category", historicCategory.getName());
		assertEquals("The description", historicCategory.getDescription());

		// this assertion holds and does not produce an exception,
		// even though there is also a map association involved!
		// But this map does not use a composite-element tag!
		assertEquals("The Text", historicCategory.getText(item));

		// set the following variable to true to demonstrate the problem
		// in case a mapping with "composite-element" tag is used.
		boolean showEnversError = true;
		
		/*
		 * If you set showEnversError to "true", the following code will produce an
		 * 
		 * org.hibernate.QueryException: could not resolve property: originalId.rev_type of: enverstest.model.Item_r
		 * 
		 * if the property "categoryItem" is mapped with the "composite-element" tag.
		 * If the property is mapped with the "element" tag using its own CompositeUserType 
		 * implementation no exception occurs!
		 * 
		 * The query produced by Envers and rejected by Hibernate is:
		 * 
		 * select new list(ee__, e__) 
		 *   from category_item_r ee__, 
		 *        enverstest.model.Item_r e__ 
		 *  where ee__.originalId.mapkey_id = e__.originalId.id 
		 *    and ee__.originalId.Category_id = :Category_id 
		 *    and e__.originalId.rev_begin.id <= :revision 
		 *    and ee__.originalId.rev_begin.id <= :revision 
		 *    and ee__.originalId.rev_type != :delrevisiontype 
		 *    and e__.originalId.rev_type != :delrevisiontype 
		 *    and (e__.rev_end.id > :revision or e__.rev_end is null) 
		 *    and (ee__.rev_end.id > :revision or ee__.rev_end is null)
		 */
		if (showEnversError) {
			Value historicValue = historicCategory.getValue(item);
			assertNotNull(historicValue);
			assertEquals("The Value", historicValue.getText());
			assertEquals(4711L, historicValue.getNumber().longValue());
			
			System.out.println("\n===== SUCCESS - all data is valid and no exception occurred! ===== \n");
		}
		
		JpaUtil.commitTransaction();
	}
}
